"""Constants for `Scorm Profile` xAPI profile."""

# Context extensions
CONTEXT_EXTENSION_COURSE_ID = "http://adlnet.gov/expapi/activities/course"
CONTEXT_EXTENSION_MODULE_ID = "http://adlnet.gov/expapi/activities/module"
