"""Constants for `LMS` xAPI profile."""

# Context extension IDs
CONTEXT_EXTENSION_STARTING_DATE = "https://w3id.org/xapi/lms/extensions/starting-date"
CONTEXT_EXTENSION_ENDING_DATE = "https://w3id.org/xapi/lms/extensions/ending-date"
CONTEXT_EXTENSION_ROLE = "https://w3id.org/xapi/lms/extensions/role"
