"""Constants for `AcrossX Profile` xAPI profile."""

# Activity extensions
ACTIVITY_EXTENSIONS_TYPE = "https://w3id.org/xapi/acrossx/extensions/type"

# Context extensions
CONTEXT_EXTENSION_SCHOOL_ID = "https://w3id.org/xapi/acrossx/extensions/school"
