"""Constants for `Video` xAPI profile."""

# Context extensions
CONTEXT_EXTENSION_CC_SUBTITLE_LANG = (
    "https://w3id.org/xapi/video/extensions/cc-subtitle-lang"
)
CONTEXT_EXTENSION_CC_ENABLED = "https://w3id.org/xapi/video/extensions/cc-enabled"
CONTEXT_EXTENSION_COMPLETION_THRESHOLD = (
    "https://w3id.org/xapi/video/extensions/completion-threshold"
)
CONTEXT_EXTENSION_FRAME_RATE = "https://w3id.org/xapi/video/extensions/frame-rate"
CONTEXT_EXTENSION_FULL_SCREEN = "https://w3id.org/xapi/video/extensions/full-screen"
CONTEXT_EXTENSION_LENGTH = "https://w3id.org/xapi/video/extensions/length"
CONTEXT_EXTENSION_PLAYED_SEGMENTS = (
    "https://w3id.org/xapi/video/extensions/played-segments"
)
CONTEXT_EXTENSION_QUALITY = "https://w3id.org/xapi/video/extensions/quality"
CONTEXT_EXTENSION_SCREEN_SIZE = "https://w3id.org/xapi/video/extensions/screen-size"
CONTEXT_EXTENSION_SESSION_ID = "https://w3id.org/xapi/video/extensions/session-id"
CONTEXT_EXTENSION_SPEED = "https://w3id.org/xapi/video/extensions/speed"

CONTEXT_EXTENSION_TRACK = "https://w3id.org/xapi/video/extensions/track"
CONTEXT_EXTENSION_USER_AGENT = "https://w3id.org/xapi/video/extensions/user-agent"
CONTEXT_EXTENSION_VIDEO_PLAYBACK_SIZE = (
    "https://w3id.org/xapi/video/extensions/video-playback-size"
)
CONTEXT_EXTENSION_VOLUME = "https://w3id.org/xapi/video/extensions/volume"

# Result extensions
RESULT_EXTENSION_TIME = "https://w3id.org/xapi/video/extensions/time"
RESULT_EXTENSION_TIME_FROM = "https://w3id.org/xapi/video/extensions/time-from"
RESULT_EXTENSION_TIME_TO = "https://w3id.org/xapi/video/extensions/time-to"
RESULT_EXTENSION_PROGRESS = "https://w3id.org/xapi/video/extensions/progress"
